@@ -1,9 +1,12 @@
 from __future__ import print_function, unicode_literals
 
 import os
+import subprocess
 import sys
 import textwrap
 import time
+import traceback
+from collections import OrderedDict
 from fnmatch import fnmatch
 
 import py
@@ -11,10 +14,11 @@ import pytest
 import six
 
 import tox
+import tox.session
 from tox import venv
 from tox.config import parseconfig
-from tox.result import ResultLog
-from tox.session import Reporter, Session, main
+from tox.config.parallel import ENV_VAR_KEY as PARALLEL_ENV_VAR_KEY
+from tox.reporter import update_default_reporter
 from tox.venv import CreationConfig, VirtualEnv, getdigest
 
 mark_dont_run_on_windows = pytest.mark.skipif(os.name == "nt", reason="non windows test")
@@ -47,6 +51,49 @@ def work_in_clean_dir(tmpdir):
         yield
 
 
+@pytest.fixture(autouse=True)
+def check_cwd_not_changed_by_test():
+    old = os.getcwd()
+    yield
+    new = os.getcwd()
+    if old != new:
+        pytest.fail("test changed cwd: {!r} => {!r}".format(old, new))
+
+
+@pytest.fixture(autouse=True)
+def check_os_environ_stable():
+    old = os.environ.copy()
+
+    to_clean = {
+        k: os.environ.pop(k, None)
+        for k in {PARALLEL_ENV_VAR_KEY, str("TOX_WORK_DIR"), str("PYTHONPATH")}
+    }
+
+    yield
+
+    for key, value in to_clean.items():
+        if value is not None:
+            os.environ[key] = value
+
+    new = os.environ
+    extra = {k: new[k] for k in set(new) - set(old)}
+    miss = {k: old[k] for k in set(old) - set(new)}
+    diff = {
+        "{} = {} vs {}".format(k, old[k], new[k])
+        for k in set(old) & set(new)
+        if old[k] != new[k] and not k.startswith("PYTEST_")
+    }
+    if extra or miss or diff:
+        msg = "test changed environ"
+        if extra:
+            msg += " extra {}".format(extra)
+        if miss:
+            msg += " miss {}".format(miss)
+        if diff:
+            msg += " diff {}".format(diff)
+        pytest.fail(msg)
+
+
 @pytest.fixture(name="newconfig")
 def create_new_config_file(tmpdir):
     def create_new_config_file_(args, source=None, plugins=()):
@@ -56,6 +103,7 @@ def create_new_config_file(tmpdir):
         s = textwrap.dedent(source)
         p = tmpdir.join("tox.ini")
         p.write(s)
+        tox.session.setup_reporter(args)
         with tmpdir.as_cwd():
             return parseconfig(args, plugins=plugins)
 
@@ -63,93 +111,115 @@ def create_new_config_file(tmpdir):
 
 
 @pytest.fixture
-def cmd(request, capfd, monkeypatch):
+def cmd(request, monkeypatch, capfd):
     if request.config.option.no_network:
         pytest.skip("--no-network was specified, test cannot run")
     request.addfinalizer(py.path.local().chdir)
 
     def run(*argv):
-        key = str("PYTHONPATH")
-        python_paths = (i for i in (os.getcwd(), os.getenv(key)) if i)
-        monkeypatch.setenv(key, os.pathsep.join(python_paths))
-
-        with RunResult(capfd, argv) as result:
-            prev_run_command = Session.runcommand
-
-            def run_command(self):
-                result.session = self
-                return prev_run_command(self)
-
-            monkeypatch.setattr(Session, "runcommand", run_command)
+        reset_report()
+        with RunResult(argv, capfd) as result:
+            _collect_session(result)
 
+            # noinspection PyBroadException
             try:
-                main([str(x) for x in argv])
+                tox.session.main([str(x) for x in argv])
                 assert False  # this should always exist with SystemExit
             except SystemExit as exception:
                 result.ret = exception.code
             except OSError as e:
+                traceback.print_exc()
                 result.ret = e.errno
+            except Exception:
+                traceback.print_exc()
+                result.ret = 1
         return result
 
+    def _collect_session(result):
+        prev_build = tox.session.build_session
+
+        def build_session(config):
+            result.session = prev_build(config)
+            return result.session
+
+        monkeypatch.setattr(tox.session, "build_session", build_session)
+
     yield run
 
 
 class RunResult:
-    def __init__(self, capfd, args):
-        self._capfd = capfd
+    def __init__(self, args, capfd):
         self.args = args
         self.ret = None
         self.duration = None
         self.out = None
         self.err = None
         self.session = None
+        self.capfd = capfd
 
     def __enter__(self):
         self._start = time.time()
-        self._capfd.readouterr()
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.duration = time.time() - self._start
-        self.out, self.err = self._capfd.readouterr()
+        self.out, self.err = self.capfd.readouterr()
+
+    def _read(self, out, pos):
+        out.buffer.seek(pos)
+        return out.buffer.read().decode(out.encoding, errors=out.errors)
 
     @property
     def outlines(self):
-        return self.out.splitlines()
+        out = [] if self.out is None else self.out.splitlines()
+        err = [] if self.err is None else self.err.splitlines()
+        return err + out
 
     def __repr__(self):
-        return "RunResult(ret={}, args={}, out=\n{}\n, err=\n{})".format(
-            self.ret, " ".join(str(i) for i in self.args), self.out, self.err
+        res = "RunResult(ret={}, args={!r}, out=\n{}\n, err=\n{})".format(
+            self.ret, self.args, self.out, self.err
         )
+        if six.PY2:
+            return res.encode("UTF-8")
+        else:
+            return res
+
+    def output(self):
+        return "{}\n{}\n{}".format(self.ret, self.err, self.out)
+
+    def assert_success(self, is_run_test_env=True):
+        msg = self.output()
+        assert self.ret == 0, msg
+        if is_run_test_env:
+            assert any("  congratulations :)" == l for l in reversed(self.outlines)), msg
+
+    def assert_fail(self, is_run_test_env=True):
+        msg = self.output()
+        assert self.ret, msg
+        if is_run_test_env:
+            assert not any("  congratulations :)" == l for l in reversed(self.outlines)), msg
 
 
 class ReportExpectMock:
-    def __init__(self, session):
-        self._calls = []
+    def __init__(self):
+        from tox import reporter
+
+        self.instance = reporter._INSTANCE
+        self.clear()
         self._index = -1
-        self.session = session
-        self.orig_reporter = Reporter(session)
 
     def clear(self):
-        self._calls[:] = []
-
-    def __getattr__(self, name):
-        if name[0] == "_":
-            raise AttributeError(name)
-        elif name == "verbosity":
-            return self.orig_reporter.verbosity
-
-        def generic_report(*args, **_):
-            self._calls.append((name,) + args)
-            print("{}".format(self._calls[-1]))
-
-        return generic_report
+        self._index = -1
+        if six.PY3:
+            self.instance.reported_lines.clear()
+        else:
+            del self.instance.reported_lines[:]
 
     def getnext(self, cat):
         __tracebackhide__ = True
         newindex = self._index + 1
-        while newindex < len(self._calls):
-            call = self._calls[newindex]
+        while newindex < len(self.instance.reported_lines):
+            call = self.instance.reported_lines[newindex]
             lcat = call[0]
             if fnmatch(lcat, cat):
                 self._index = newindex
@@ -157,7 +227,7 @@ class ReportExpectMock:
             newindex += 1
         raise LookupError(
             "looking for {!r}, no reports found at >={:d} in {!r}".format(
-                cat, self._index + 1, self._calls
+                cat, self._index + 1, self.instance.reported_lines
             )
         )
 
@@ -165,7 +235,7 @@ class ReportExpectMock:
         __tracebackhide__ = True
         if not messagepattern.startswith("*"):
             messagepattern = "*{}".format(messagepattern)
-        while self._index < len(self._calls):
+        while self._index < len(self.instance.reported_lines):
             try:
                 call = self.getnext(cat)
             except LookupError:
@@ -181,7 +251,7 @@ class ReportExpectMock:
         if not invert:
             raise AssertionError(
                 "looking for {}({!r}), no reports found at >={:d} in {!r}".format(
-                    cat, messagepattern, self._index + 1, self._calls
+                    cat, messagepattern, self._index + 1, self.instance.reported_lines
                 )
             )
 
@@ -192,12 +262,14 @@ class ReportExpectMock:
 class pcallMock:
     def __init__(self, args, cwd, env, stdout, stderr, shell):
         self.arg0 = args[0]
-        self.args = args[1:]
+        self.args = args
         self.cwd = cwd
         self.env = env
         self.stdout = stdout
         self.stderr = stderr
         self.shell = shell
+        self.pid = os.getpid()
+        self.returncode = 0
 
     @staticmethod
     def communicate():
@@ -209,36 +281,47 @@ class pcallMock:
 
 @pytest.fixture(name="mocksession")
 def create_mocksession(request):
-    class MockSession(Session):
-        def __init__(self):
-            self._clearmocks()
-            self.config = request.getfixturevalue("newconfig")([], "")
-            self.resultlog = ResultLog()
-            self._actions = []
+    config = request.getfixturevalue("newconfig")([], "")
 
-        def getenv(self, name):
-            return VirtualEnv(self.config.envconfigs[name], session=self)
+    class MockSession(tox.session.Session):
+        def __init__(self, config):
+            self.logging_levels(config.option.quiet_level, config.option.verbose_level)
+            super(MockSession, self).__init__(config, popen=self.popen)
+            self._pcalls = []
+            self.report = ReportExpectMock()
 
         def _clearmocks(self):
-            self._pcalls = []
-            self._spec2pkg = {}
-            self.report = ReportExpectMock(self)
-
-        def make_emptydir(self, path):
-            pass
+            if six.PY3:
+                self._pcalls.clear()
+            else:
+                del self._pcalls[:]
+            self.report.clear()
 
         def popen(self, args, cwd, shell=None, stdout=None, stderr=None, env=None, **_):
-            pm = pcallMock(args, cwd, env, stdout, stderr, shell)
-            self._pcalls.append(pm)
-            return pm
+            process_call_mock = pcallMock(args, cwd, env, stdout, stderr, shell)
+            self._pcalls.append(process_call_mock)
+            return process_call_mock
 
-    return MockSession()
+        def new_config(self, config):
+            self.logging_levels(config.option.quiet_level, config.option.verbose_level)
+            self.config = config
+            self.venv_dict.clear()
+            self.existing_venvs.clear()
+
+        def logging_levels(self, quiet, verbose):
+            update_default_reporter(quiet, verbose)
+            if hasattr(self, "config"):
+                self.config.option.quiet_level = quiet
+                self.config.option.verbose_level = verbose
+
+    return MockSession(config)
 
 
 @pytest.fixture
 def newmocksession(mocksession, newconfig):
     def newmocksession_(args, source, plugins=()):
-        mocksession.config = newconfig(args, source, plugins=plugins)
+        config = newconfig(args, source, plugins=plugins)
+        mocksession._reset(config, mocksession.popen)
         return mocksession
 
     return newmocksession_
@@ -283,7 +366,7 @@ def initproj(tmpdir):
         if not src_root:
             src_root = "."
         if isinstance(nameversion, six.string_types):
-            parts = nameversion.split(str("-"))
+            parts = nameversion.rsplit(str("-"), 1)
             if len(parts) == 1:
                 parts.append("0.1")
             name, version = parts
@@ -336,11 +419,11 @@ def initproj(tmpdir):
             "include {}".format(p.relto(base)) for p in base.visit(lambda x: x.check(file=1))
         ]
         create_files(base, {"MANIFEST.in": "\n".join(manifestlines)})
-        print("created project in {}".format(base))
         base.chdir()
         return base
 
-    return initproj_
+    with py.path.local().as_cwd():
+        yield initproj_
 
 
 def _path_parts(path):
@@ -404,16 +487,15 @@ def mock_venv(monkeypatch):
     and cannot install any packages. """
 
     # first ensure we have a clean python path
-    monkeypatch.delenv("PYTHONPATH", raising=False)
+    monkeypatch.delenv(str("PYTHONPATH"), raising=False)
 
     # object to collect some data during the execution
     class Result(object):
-        def __init__(self):
-            self.popens = []
-            self.cwd = None
-            self.session = None
+        def __init__(self, session):
+            self.popens = popen_list
+            self.session = session
 
-    res = Result()
+    res = OrderedDict()
 
     # convince tox that the current running virtual environment is already the env we would create
     class ProxyCurrentPython:
@@ -465,31 +547,29 @@ def mock_venv(monkeypatch):
     monkeypatch.setattr(venv, "tox_runenvreport", tox_runenvreport)
 
     # intercept the build session to save it and we intercept the popen invocations
-    prev_build = tox.session.build_session
-
-    def build_session(config):
-        res.session = prev_build(config)
-        res._popen = res.session.popen
-        monkeypatch.setattr(res.session, "popen", popen)
-        return res.session
-
-    monkeypatch.setattr(tox.session, "build_session", build_session)
-
     # collect all popen calls
+    popen_list = []
+
     def popen(cmd, **kwargs):
-        # we don't want to perform installation of new packages, just replace with an always ok cmd
+        # we don't want to perform installation of new packages,
+        # just replace with an always ok cmd
         if "pip" in cmd and "install" in cmd:
             cmd = ["python", "-c", "print({!r})".format(cmd)]
-        activity_id = res.session._actions[-1].id
-        activity_name = res.session._actions[-1].activity
+        ret = None
         try:
-            ret = res._popen(cmd, **kwargs)
+            ret = subprocess.Popen(cmd, **kwargs)
         except tox.exception.InvocationError as exception:  # pragma: no cover
             ret = exception  # pragma: no cover
         finally:
-            res.popens.append((activity_id, activity_name, kwargs.get("env"), ret, cmd))
+            popen_list.append((kwargs.get("env"), ret, cmd))
         return ret
 
+    def build_session(config):
+        session = tox.session.Session(config, popen=popen)
+        res[id(session)] = Result(session)
+        return session
+
+    monkeypatch.setattr(tox.session, "build_session", build_session)
     return res
 
 
@@ -498,3 +578,21 @@ def current_tox_py():
     """generate the current (test runners) python versions key
     e.g. py37 when running under Python 3.7"""
     return "py{}".format("".join(str(i) for i in sys.version_info[0:2]))
+
+
+def pytest_runtest_setup(item):
+    reset_report()
+
+
+def pytest_runtest_teardown(item):
+    reset_report()
+
+
+def pytest_pyfunc_call(pyfuncitem):
+    reset_report()
+
+
+def reset_report(quiet=0, verbose=0):
+    from tox.reporter import _INSTANCE
+
+    _INSTANCE._reset(quiet_level=quiet, verbose_level=verbose)
