@@ -21,7 +21,17 @@ import toml
 import tox
 from tox.constants import INFO
 from tox.interpreters import Interpreters, NoInterpreterInfo
-from .parallel import add_parallel_flags, ENV_VAR_KEY as PARALLEL_ENV_VAR_KEY, add_parallel_config
+from tox.reporter import (
+    REPORTER_TIMESTAMP_ON_ENV,
+    error,
+    update_default_reporter,
+    using,
+    verbosity1,
+)
+
+from .parallel import ENV_VAR_KEY as PARALLEL_ENV_VAR_KEY
+from .parallel import add_parallel_config, add_parallel_flags
+from .reporter import add_verbosity_commands
 
 hookimpl = tox.hookimpl
 """DEPRECATED - REMOVE - this is left for compatibility with plugins importing this from here.
@@ -148,6 +158,7 @@ class DepOption:
                     name_start = "{} ".format(option)
                     if name.startswith(name_start):
                         name = "{}={}".format(option, name[len(option) :].strip())
+            name = self._cut_off_dep_comment(name)
             name = self._replace_forced_dep(name, config)
             deps.append(DepConfig(name, ixserver))
         return deps
@@ -166,6 +177,10 @@ class DepOption:
                 return forced_dep
         return name
 
+    @staticmethod
+    def _cut_off_dep_comment(name):
+        return re.sub(r"\s+#.*", "", name).strip()
+
     @classmethod
     def _is_same_dep(cls, dep1, dep2):
         """Definitions are the same if they refer to the same package, even if versions differ."""
@@ -224,6 +239,7 @@ def parseconfig(args, plugins=()):
     """
     pm = get_plugin_manager(plugins)
     config, option = parse_cli(args, pm)
+    update_default_reporter(config.option.quiet_level, config.option.verbose_level)
 
     for config_file in propose_configs(option.configfile):
         config_type = config_file.basename
@@ -364,22 +380,7 @@ def tox_addoption(parser):
     parser.add_argument(
         "--help-ini", "--hi", action="store_true", dest="helpini", help="show help about ini-names"
     )
-    parser.add_argument(
-        "-v",
-        action="count",
-        dest="verbose_level",
-        default=0,
-        help="increase verbosity of reporting output."
-        "-vv mode turns off output redirection for package installation, "
-        "above level two verbosity flags are passed through to pip (with two less level)",
-    )
-    parser.add_argument(
-        "-q",
-        action="count",
-        dest="quiet_level",
-        default=0,
-        help="progressively silence reporting output.",
-    )
+    add_verbosity_commands(parser)
     parser.add_argument(
         "--showconfig",
         action="store_true",
@@ -565,25 +566,28 @@ def tox_addoption(parser):
         proposed_python = (implied_python or sys.executable) if value is None else str(value)
         if implied_python is not None and implied_python != proposed_python:
             testenv_config.basepython = proposed_python
-            implied_version = tox.PYTHON.PY_FACTORS_RE.match(factor).group(2)
-            python_info_for_proposed = testenv_config.python_info
-            if not isinstance(python_info_for_proposed, NoInterpreterInfo):
-                proposed_version = "".join(
-                    str(i) for i in python_info_for_proposed.version_info[0:2]
-                )
-                if implied_version != proposed_version:
-                    # TODO(stephenfin): Raise an exception here in tox 4.0
-                    warnings.warn(
-                        "conflicting basepython version (set {}, should be {}) for env '{}';"
-                        "resolve conflict or set ignore_basepython_conflict".format(
-                            proposed_version, implied_version, testenv_config.envname
-                        )
+            match = tox.PYTHON.PY_FACTORS_RE.match(factor)
+            implied_version = match.group(2) if match else None
+            if implied_version is not None:
+                python_info_for_proposed = testenv_config.python_info
+                if not isinstance(python_info_for_proposed, NoInterpreterInfo):
+                    proposed_version = "".join(
+                        str(i) for i in python_info_for_proposed.version_info[0:2]
                     )
+                    # '27'.startswith('2') or '27'.startswith('27')
+                    if not proposed_version.startswith(implied_version):
+                        # TODO(stephenfin): Raise an exception here in tox 4.0
+                        warnings.warn(
+                            "conflicting basepython version (set {}, should be {}) for env '{}';"
+                            "resolve conflict or set ignore_basepython_conflict".format(
+                                proposed_version, implied_version, testenv_config.envname
+                            )
+                        )
         return proposed_python
 
     parser.add_testenv_attribute(
         name="basepython",
-        type="string",
+        type="basepython",
         default=None,
         postprocess=basepython_default,
         help="executable name or path of interpreter used to create a virtual test environment.",
@@ -659,7 +663,16 @@ def tox_addoption(parser):
         # Flatten the list to deal with space-separated values.
         value = list(itertools.chain.from_iterable([x.split(" ") for x in value]))
 
-        passenv = {"PATH", "PIP_INDEX_URL", "LANG", "LANGUAGE", "LD_LIBRARY_PATH", "TOX_WORK_DIR"}
+        passenv = {
+            "PATH",
+            "PIP_INDEX_URL",
+            "LANG",
+            "LANGUAGE",
+            "LD_LIBRARY_PATH",
+            "TOX_WORK_DIR",
+            str(REPORTER_TIMESTAMP_ON_ENV),
+            str(PARALLEL_ENV_VAR_KEY),
+        }
 
         # read in global passenv settings
         p = os.environ.get("TOX_TESTENV_PASSENV", None)
@@ -950,9 +963,19 @@ def make_hashseed():
 class ParseIni(object):
     def __init__(self, config, ini_path, ini_data):  # noqa
         config.toxinipath = ini_path
+        using("tox.ini: {} (pid {})".format(config.toxinipath, os.getpid()))
         config.toxinidir = config.toxinipath.dirpath()
 
         self._cfg = py.iniconfig.IniConfig(config.toxinipath, ini_data)
+        previous_line_of = self._cfg.lineof
+
+        def line_of_default_to_zero(section, name=None):
+            at = previous_line_of(section, name=name)
+            if at is None:
+                at = 0
+            return at
+
+        self._cfg.lineof = line_of_default_to_zero
         config._cfg = self._cfg
         self.config = config
 
@@ -979,32 +1002,28 @@ class ParseIni(object):
         config.hashseed = hash_seed
 
         reader.addsubstitutions(toxinidir=config.toxinidir, homedir=config.homedir)
-        # As older versions of tox may have bugs or incompatibilities that
-        # prevent parsing of tox.ini this must be the first thing checked.
-        config.minversion = reader.getstring("minversion", None)
-        if config.minversion:
-            tox_version = pkg_resources.parse_version(tox.__version__)
-            config_min_version = pkg_resources.parse_version(self.config.minversion)
-            if config_min_version > tox_version:
-                raise tox.exception.MinVersionError(
-                    "tox version is {}, required is at least {}".format(
-                        tox.__version__, self.config.minversion
-                    )
-                )
-
-        self.ensure_requires_satisfied(reader.getlist("requires"))
 
         if config.option.workdir is None:
             config.toxworkdir = reader.getpath("toxworkdir", "{toxinidir}/.tox")
         else:
             config.toxworkdir = config.toxinidir.join(config.option.workdir, abs=True)
 
-        if config.option.skip_missing_interpreters == "config":
-            val = reader.getbool("skip_missing_interpreters", False)
-            config.option.skip_missing_interpreters = "true" if val else "false"
+        if os.path.exists(str(config.toxworkdir)):
+            config.toxworkdir = config.toxworkdir.realpath()
 
+        reader.addsubstitutions(toxworkdir=config.toxworkdir)
         config.ignore_basepython_conflict = reader.getbool("ignore_basepython_conflict", False)
 
+        config.distdir = reader.getpath("distdir", "{toxworkdir}/dist")
+
+        reader.addsubstitutions(distdir=config.distdir)
+        config.distshare = reader.getpath("distshare", dist_share_default)
+        config.temp_dir = reader.getpath("temp_dir", "{toxworkdir}/.tmp")
+        reader.addsubstitutions(distshare=config.distshare)
+        config.sdistsrc = reader.getpath("sdistsrc", None)
+        config.setupdir = reader.getpath("setupdir", "{toxinidir}")
+        config.logdir = config.toxworkdir.join("log")
+
         # determine indexserver dictionary
         config.indexserver = {"default": IndexServerConfig("default")}
         prefix = "indexserver"
@@ -1012,6 +1031,10 @@ class ParseIni(object):
             name, url = map(lambda x: x.strip(), line.split("=", 1))
             config.indexserver[name] = IndexServerConfig(name, url)
 
+        if config.option.skip_missing_interpreters == "config":
+            val = reader.getbool("skip_missing_interpreters", False)
+            config.option.skip_missing_interpreters = "true" if val else "false"
+
         override = False
         if config.option.indexurl:
             for url_def in config.option.indexurl:
@@ -1032,16 +1055,7 @@ class ParseIni(object):
             for name in config.indexserver:
                 config.indexserver[name] = IndexServerConfig(name, override)
 
-        reader.addsubstitutions(toxworkdir=config.toxworkdir)
-        config.distdir = reader.getpath("distdir", "{toxworkdir}/dist")
-
-        reader.addsubstitutions(distdir=config.distdir)
-        config.distshare = reader.getpath("distshare", dist_share_default)
-        config.temp_dir = reader.getpath("temp_dir", "{toxworkdir}/.tmp")
-        reader.addsubstitutions(distshare=config.distshare)
-        config.sdistsrc = reader.getpath("sdistsrc", None)
-        config.setupdir = reader.getpath("setupdir", "{toxinidir}")
-        config.logdir = config.toxworkdir.join("log")
+        self.handle_provision(config, reader)
 
         self.parse_build_isolation(config, reader)
         config.envlist, all_envs = self._getenvdata(reader, config)
@@ -1053,7 +1067,7 @@ class ParseIni(object):
         # factors stated in config envlist
         stated_envlist = reader.getstring("envlist", replace=False)
         if stated_envlist:
-            for env in config.envlist:
+            for env in _split_env(stated_envlist):
                 known_factors.update(env.split("-"))
 
         # configure testenvs
@@ -1068,7 +1082,6 @@ class ParseIni(object):
                 )
             ):
                 config.envconfigs[name] = self.make_envconfig(name, section, reader._subs, config)
-
         all_develop = all(
             name in config.envconfigs and config.envconfigs[name].usedevelop
             for name in config.envlist
@@ -1076,31 +1089,62 @@ class ParseIni(object):
 
         config.skipsdist = reader.getbool("skipsdist", all_develop)
 
+    def handle_provision(self, config, reader):
+        requires_list = reader.getlist("requires")
+        config.minversion = reader.getstring("minversion", None)
+        config.provision_tox_env = name = reader.getstring("provision_tox_env", ".tox")
+        min_version = "tox >= {}".format(config.minversion or tox.__version__)
+        deps = self.ensure_requires_satisfied(config, requires_list, min_version)
+        if config.run_provision:
+            section_name = "testenv:{}".format(name)
+            if section_name not in self._cfg.sections:
+                self._cfg.sections[section_name] = {}
+            self._cfg.sections[section_name]["description"] = "meta tox"
+            env_config = self.make_envconfig(
+                name, "{}{}".format(testenvprefix, name), reader._subs, config
+            )
+            env_config.deps = deps
+            config.envconfigs[config.provision_tox_env] = env_config
+            raise tox.exception.MissingRequirement(config)
+
+    @staticmethod
+    def ensure_requires_satisfied(config, requires, min_version):
+        missing_requirements = []
+        failed_to_parse = False
+        deps = []
+        exists = set()
+        for require in requires + [min_version]:
+            # noinspection PyBroadException
+            try:
+                package = pkg_resources.Requirement.parse(require)
+                if package.project_name not in exists:
+                    deps.append(DepConfig(require, None))
+                    exists.add(package.project_name)
+                    pkg_resources.get_distribution(package)
+            except pkg_resources.RequirementParseError as exception:
+                failed_to_parse = True
+                error("failed to parse {!r}".format(exception))
+            except Exception as exception:
+                verbosity1("could not satisfy requires {!r}".format(exception))
+                missing_requirements.append(str(pkg_resources.Requirement(require)))
+        if failed_to_parse:
+            raise tox.exception.BadRequirement()
+        config.run_provision = bool(len(missing_requirements))
+        return deps
+
     def parse_build_isolation(self, config, reader):
         config.isolated_build = reader.getbool("isolated_build", False)
         config.isolated_build_env = reader.getstring("isolated_build_env", ".package")
         if config.isolated_build is True:
             name = config.isolated_build_env
-            if name not in config.envconfigs:
-                config.envconfigs[name] = self.make_envconfig(
-                    name, "{}{}".format(testenvprefix, name), reader._subs, config
-                )
-
-    @staticmethod
-    def ensure_requires_satisfied(specified):
-        missing_requirements = []
-        for s in specified:
-            try:
-                pkg_resources.get_distribution(s)
-            except pkg_resources.RequirementParseError:
-                raise
-            except Exception:
-                missing_requirements.append(str(pkg_resources.Requirement(s)))
-        if missing_requirements:
-            raise tox.exception.MissingRequirement(
-                "Packages {} need to be installed alongside tox in {}".format(
-                    ", ".join(missing_requirements), sys.executable
-                )
+            section_name = "testenv:{}".format(name)
+            if section_name not in self._cfg.sections:
+                self._cfg.sections[section_name] = {}
+            self._cfg.sections[section_name]["deps"] = ""
+            self._cfg.sections[section_name]["sitepackages"] = "False"
+            self._cfg.sections[section_name]["description"] = "isolated packaging environment"
+            config.envconfigs[name] = self.make_envconfig(
+                name, "{}{}".format(testenvprefix, name), reader._subs, config
             )
 
     def _list_section_factors(self, section):
@@ -1128,6 +1172,11 @@ class ParseIni(object):
                 if atype in ("bool", "path", "string", "dict", "dict_setenv", "argv", "argvlist"):
                     meth = getattr(reader, "get{}".format(atype))
                     res = meth(env_attr.name, env_attr.default, replace=replace)
+                elif atype == "basepython":
+                    no_fallback = name in (config.provision_tox_env,)
+                    res = reader.getstring(
+                        env_attr.name, env_attr.default, replace=replace, no_fallback=no_fallback
+                    )
                 elif atype == "space-separated-list":
                     res = reader.getlist(env_attr.name, sep=" ")
                 elif atype == "line-list":
@@ -1143,43 +1192,58 @@ class ParseIni(object):
                 tc.missing_subs.append(e.name)
                 res = e.FLAG
             setattr(tc, env_attr.name, res)
-            if atype in ("path", "string"):
+            if atype in ("path", "string", "basepython"):
                 reader.addsubstitutions(**{env_attr.name: res})
         return tc
 
-    def _getenvdata(self, reader, config):
-        candidates = (
-            os.environ.get(PARALLEL_ENV_VAR_KEY),
-            self.config.option.env,
-            os.environ.get("TOXENV"),
-            reader.getstring("envlist", replace=False),
-        )
-        env_str = next((i for i in candidates if i), [])
+    def _getallenvs(self, reader, extra_env_list=None):
+        extra_env_list = extra_env_list or []
+        env_str = reader.getstring("envlist", replace=False)
         env_list = _split_env(env_str)
+        for env in extra_env_list:
+            if env not in env_list:
+                env_list.append(env)
 
-        # collect section envs
         all_envs = OrderedDict((i, None) for i in env_list)
-        if "ALL" in all_envs:
-            all_envs.pop("ALL")
         for section in self._cfg:
             if section.name.startswith(testenvprefix):
                 all_envs[section.name[len(testenvprefix) :]] = None
         if not all_envs:
             all_envs["python"] = None
+        return list(all_envs.keys())
+
+    def _getenvdata(self, reader, config):
+        from_option = self.config.option.env
+        from_environ = os.environ.get("TOXENV")
+        from_config = reader.getstring("envlist", replace=False)
+
+        env_list = []
+        if (from_option and "ALL" in from_option) or (
+            not from_option and from_environ and "ALL" in from_environ.split(",")
+        ):
+            all_envs = self._getallenvs(reader)
+        else:
+            candidates = (
+                os.environ.get(PARALLEL_ENV_VAR_KEY),
+                from_option,
+                from_environ,
+                from_config,
+            )
+            env_str = next((i for i in candidates if i), [])
+            env_list = _split_env(env_str)
+            all_envs = self._getallenvs(reader, env_list)
+
+        if not env_list:
+            env_list = all_envs
 
         package_env = config.isolated_build_env
         if config.isolated_build is True and package_env in all_envs:
-            all_envs.pop(package_env)
-
-        if not env_list or "ALL" in env_list:
-            env_list = list(all_envs.keys())
+            all_envs.remove(package_env)
 
         if config.isolated_build is True and package_env in env_list:
             msg = "isolated_build_env {} cannot be part of envlist".format(package_env)
             raise tox.exception.ConfigError(msg)
-
-        all_env_list = list(all_envs.keys())
-        return env_list, all_env_list
+        return env_list, all_envs
 
 
 def _split_env(env):
@@ -1344,9 +1408,10 @@ class SectionReader:
     def getargv(self, name, default="", replace=True):
         return self.getargvlist(name, default, replace=replace)[0]
 
-    def getstring(self, name, default=None, replace=True, crossonly=False):
+    def getstring(self, name, default=None, replace=True, crossonly=False, no_fallback=False):
         x = None
-        for s in [self.section_name] + self.fallbacksections:
+        sections = [self.section_name] + ([] if no_fallback else self.fallbacksections)
+        for s in sections:
             try:
                 x = self._cfg[s][name]
                 break
